//         Ŀ
//          Keyboard handler for DJGPP version of diskmag Vyhen 
//                         code: HH/AA (C) 1997                 
//         

#include <stdio.h>
#include <conio.h>
#include <go32.h>
#include <dpmi.h>
#include <sys/nearptr.h>
#include <dos.h>
#include <dpmi.h>
#include <sys/movedata.h>

#define FALSE 0
#define TRUE 1
#define END_OF_FUNCTION(x)    void x##_end() { }
#define LOCK_VARIABLE(x)      _go32_dpmi_lock_data((void *)&x, sizeof(x))
#define LOCK_FUNCTION(x)      _go32_dpmi_lock_code(x, (long)x##_end - (long)x)

#define DISABLE()   asm volatile ("cli")
#define ENABLE()    asm volatile ("sti")

#define SCANCODE_TO_ALT(c)       (0)

#define KB_SHIFT_FLAG         1
#define KB_CTRL_FLAG          2
#define KB_ALT_FLAG           4
#define KB_LWIN_FLAG          8
#define KB_RWIN_FLAG          16
#define KB_MENU_FLAG          32
#define KB_SCROLOCK_FLAG      256
#define KB_NUMLOCK_FLAG       512
#define KB_CAPSLOCK_FLAG      1024
#define KB_INALTSEQ_FLAG      4096

#define KEY_CONTROL           29
#define KEY_LCONTROL          29
#define KEY_LSHIFT            42
#define KEY_RSHIFT            54
#define KEY_ALT               56
#define KEY_CAPSLOCK          58


/* Most of code is borrowed from allegro - an grapics library for djgpp by
 *      By Shawn Hargreaves,
 *      1 Salisbury Road,
 *      Market Drayton,
 *      Shropshire,
 *      England, TF9 1AJ.
 *
 */
#define KEYBOARD_INT          9

#define KB_SPECIAL_MASK       0x3F
#define KB_CTRL_ALT_FLAG      (KB_CTRL_FLAG | KB_ALT_FLAG)


static int three_finger_flag = TRUE;
static int key_led_flag = TRUE;

static int keyboard_installed = FALSE;

volatile char v_key[128];                   /* key pressed flags */

static volatile int key_shifts = 0;



#if 0
/* kb_wait_for_write_ready:
 *  Wait for the keyboard controller to set the ready-for-write bit.
 */
static inline void kb_wait_for_write_ready()
{
   long i = clock();

   while (i-100<clock() && (inportb(0x64) & 2))
      ; /* wait */
}



/* kb_wait_for_read_ready:
 *  Wait for the keyboard controller to set the ready-for-read bit.
 */
static inline void kb_wait_for_read_ready()
{
   long i = clock();

   while (i-100<clock() && (!(inportb(0x64) & 0x01)))
      ; /* wait */
}



/* kb_send_data:
 *  Sends a byte to the keyboard controller. Returns 1 if all OK.
 */
static inline int kb_send_data(unsigned char data)
{
   long i;
   int resends = 4;
   int temp;

   do {
      kb_wait_for_write_ready();

      outportb(0x60, data);
      i = 2000000L;

      while (--i) {
         kb_wait_for_read_ready();
         temp = inportb(0x60);

         if (temp == 0xFA)
            return 1;
         else if (temp == 0xFE)
            break;
      }
   }
   while ((resends-- > 0) && (i));

   return 0;
}



/* update_leds:
 *  Sets the state of the keyboard LED indicators.
 */
static inline void update_leds()
{
   if ((!kb_send_data(0xED)) || (!kb_send_data((key_shifts>>8) & 7)))
      kb_send_data(0xF4);
}
#endif



/* my_keyint:
 *  Hardware level keyboard interrupt (int 9) handler.
 */
static int my_keyint()
{
   int  temp,tmp1, release;
   static volatile int fshift;

   tmp1=temp = inportb(0x60);            /* read keyboard byte */
    release = (temp & 0x80);      /* bit 7 means key was released */
    temp &= 0x7F;                 /* bits 0-6 is the scan code */
    if (release) {                /* key was released */
     v_key[temp] = FALSE;
    }
    else {                        /* key was pressed */
     if(temp!=42||!fshift) {
     v_key[temp] = TRUE;
     }
    }
   fshift=0;
   if((char)tmp1==-32) fshift=1;
   outportb(0x20,0x20);       /* ack. the interrupt */
   return 0;
}

static END_OF_FUNCTION(my_keyint);



/* install_keyboard:
 *  Installs Allegro's keyboard handler. You must call this before using
 *  any of the keyboard input routines. Note that Allegro completely takes
 *  over the keyboard, so the debugger will not work properly, and under
 *  DOS even ctrl-alt-del will have no effect. Returns -1 on failure.
 */
static _go32_dpmi_seginfo pmint,oldint;
int v_install_keyboard()
{
   unsigned short shifts;

   if (keyboard_installed)
      return -1;

   LOCK_VARIABLE(three_finger_flag);
   LOCK_VARIABLE(key_led_flag);
   LOCK_VARIABLE(key_shifts);
   LOCK_VARIABLE(v_key);
   LOCK_FUNCTION(my_keyint);

   while (kbhit())
      getch();

   /* get state info from the BIOS */
   _dosmemgetw(0x417, 1, &shifts);
   memset((void *)v_key,0,128);
/*   printf("%i\n",shifts);
   sleep(5);*/

   key_shifts = 0;

   if (shifts & 1) {
      key_shifts |= KB_SHIFT_FLAG;
      v_key[KEY_RSHIFT] = TRUE;
   }
   if (shifts & 2) {
      key_shifts |= KB_SHIFT_FLAG;
      v_key[KEY_LSHIFT] = TRUE;
   }
   if (shifts & 4) {
      key_shifts |= KB_CTRL_FLAG;
      v_key[KEY_LCONTROL] = TRUE;
   }
   if (shifts & 8) {
      key_shifts |= KB_ALT_FLAG;
      v_key[KEY_ALT] = TRUE;
   }
   if (shifts & 16)
      key_shifts |= KB_SCROLOCK_FLAG;
   if (shifts & 32)
      key_shifts |= KB_NUMLOCK_FLAG;
   if (shifts & 64)
      key_shifts |= KB_CAPSLOCK_FLAG;



   pmint.pm_selector = _my_cs();
   pmint.pm_offset = (unsigned)&my_keyint;
   _go32_dpmi_allocate_iret_wrapper(&pmint);
   DISABLE();
   _go32_dpmi_get_protected_mode_interrupt_vector(9, &oldint);
   _go32_dpmi_set_protected_mode_interrupt_vector(9, &pmint);

   /*update_leds();*/
   ENABLE();

   /*_add_exit_func(remove_keyboard);*/
   keyboard_installed = TRUE;
   return 0;
}



/* remove_keyboard:
 *  Removes the keyboard handler, returning control to the BIOS. You don't
 *  normally need to call this, because allegro_exit() will do it for you.
 */
void v_remove_keyboard()
{
   unsigned short shifts;

   if (!keyboard_installed)
      return;

   /*_remove_irq(KEYBOARD_INT);*/

   /* transfer state info to the BIOS */
   shifts = 0;

   if (v_key[KEY_RSHIFT])
      shifts |= 1;
   if (v_key[KEY_LSHIFT])
      shifts |= 2;
   if (v_key[KEY_LCONTROL])
      shifts |= 4;
   if (v_key[KEY_ALT])
      shifts |= 8;
   if (key_shifts & KB_SCROLOCK_FLAG)
      shifts |= 16;
   if (key_shifts & KB_NUMLOCK_FLAG)
      shifts |= 32;
   if (key_shifts & KB_CAPSLOCK_FLAG)
      shifts |= 64;

   _dosmemputw(&shifts, 1, 0x417);

   /*_remove_exit_func(remove_keyboard);*/
   _go32_dpmi_set_protected_mode_interrupt_vector(9, &oldint);
   _go32_dpmi_free_iret_wrapper(&pmint);
   keyboard_installed = FALSE;
}



